import json
from avogadro.core import Molecule
from avogadro.io import FileFormatManager
from .base import BaseReader

class NWChemJsonReader(BaseReader):

    def read(self):
        str_data = self._file.read()
        mol = Molecule()
        conv = FileFormatManager()
        conv.readString(mol, str_data, 'json')
        cjson_str = conv.writeString(mol, 'cjson')
        cjson = json.loads(cjson_str)
        return cjson



from bson.objectid import ObjectId
import cumulus
from cumulus.taskflow import TaskFlow
from cumulus.taskflow.cluster import create_girder_client
from cumulus.tasks.job import (download_job_input_folders,
                               upload_job_output_to_folder)
from cumulus.tasks.job import submit_job, monitor_job

from girder.api.rest import getCurrentUser
from girder.constants import AccessType
from girder.utility.model_importer import ModelImporter

from jsonpath_rw import parse
import os
import datetime
import json
from io import BytesIO
import tempfile

from abc import ABC, abstractmethod

class OpenChemistryTaskFlow(TaskFlow, ABC):
    """
    {
        "input": {
            "calculation": {
                "_id": <the id of the pending calculation>
            },
        },
        "cluster": {
            "_id": <id of cluster to run on>
        }
    }
    """

    def start(self, *args, **kwargs):
        user = getCurrentUser()
        input_ = kwargs.get('input')
        cluster = kwargs.get('cluster')

        if input_ is None:
            raise Exception('Unable to extract input.')

        if '_id' not in cluster and 'name' not in cluster:
            raise Exception('Unable to extract cluster.')

        cluster_id = parse('cluster._id').find(kwargs)
        if cluster_id:
            cluster_id = cluster_id[0].value
            model = ModelImporter.model('cluster', 'cumulus')
            cluster = model.load(cluster_id, user=user, level=AccessType.ADMIN)
            cluster = model.filter(cluster, user, passphrase=False)

        super(OpenChemistryTaskFlow, self).start(
            setup_input_template.s(input_, cluster),
            *args, **kwargs)

    @property
    @abstractmethod
    def code_label(self):
        pass

    @property
    @abstractmethod
    def docker_image(self):
        pass

    @abstractmethod
    def input_generator(self, params, cjson, tmp_file):
        pass

    @abstractmethod
    def select_output_files(self, filenames):
        pass

    def ec2_job_commands(self, input_name):
        return self.demo_job_commands(input_name)

    def demo_job_commands(self, input_name):
        mount_dir = '/data'
        return [
            'docker pull %s' % self.docker_image,
            'docker run -v dev_job_data:%s %s %s' % (
                mount_dir,
                self.docker_image,
                os.path.join(mount_dir, '{{job._id}}', input_name)
            )
        ]

    def nersc_job_commands(self, input_name):
        raise NotImplementedError('%s has not been configured to run on NERSC yet.' % self.code_label)

def _get_cori(client):
    params = {
        'type': 'newt'
    }
    clusters = client.get('clusters', parameters=params)
    for cluster in clusters:
        if cluster['name'] == 'cori':
            return cluster

    # We need to create one
    body = {
        'config': {
            'host': 'cori'
        },
        'name': 'cori',
        'type': 'newt'
    }
    cluster = client.post('clusters', data=json.dumps(body))

    return cluster

def _get_oc_folder(client):
    me = client.get('user/me')
    if me is None:
        raise Exception('Unable to get me.')

    login = me['login']
    private_folder_path =    'user/%s/Private' % login
    private_folder = client.resourceLookup(private_folder_path)
    oc_folder_path = '%s/oc' % private_folder_path
    oc_folder = client.resourceLookup(oc_folder_path, test=True)
    if oc_folder is None:
        oc_folder = client.createFolder(private_folder['_id'], 'oc')

    return oc_folder

def _fetch_best_geometry(client, molecule_id):
    # Fetch our best geometry
    params = {
        'moleculeId': molecule_id,
        'sortByTheory': True,
        'limit': 1,
        'calculationType': 'optimization',
        'pending': False
    }

    calculations = client.get('calculations', parameters=params)

    if len(calculations) < 1:
        return None

    return calculations[0]


@cumulus.taskflow.task
def setup_input_template(task, input_, cluster):
    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    if cluster.get('name') == 'cori':
        cluster = _get_cori(client)

    if '_id' not in cluster:
        raise Exception('Invalid cluster configurations: %s' % cluster)

    optimize = input_['optimize']
    calculation_id = parse('calculation._id').find(input_)
    if not calculation_id:
        raise Exception('Unable to extract calculation id.')
    calculation_id = calculation_id[0].value
    calculation = client.get('calculations/%s' % calculation_id)
    molecule_id = calculation['moleculeId']

    optimization_calculation_id = None
    input_calculation = parse('properties.input.calculationId').find(calculation)
    # We have been asked to use a specific calculation
    if input_calculation:
        optimization_calculation_id = input_calculation[0].value
    # We have been asked to use a specific optimized geometry, see if we have it
    elif optimize:
        parameters = {
            'moleculeId': molecule_id,
            'calculationType': 'optimizations',
        }

        basis = parse('properties.basisSet.name').find(calculation)
        if basis:
            parameters['basis'] = basis[0].value

        functional = parse('properties.functional').find(calculation)
        if functional:
            parameters['functional'] = functional[0].value.lower()

        theory = parse('properties.theory').find(calculation)
        if theory:
            parameters['theory'] = theory[0].value.lower()


        calculations = client.get('calculations', parameters)

        if len(calculations) > 0:
            optimization_calculation_id = calculations[0]['_id']

    best_calc = None
    if optimization_calculation_id is None:
        best_calc = _fetch_best_geometry(client, molecule_id)

    # We are using a specific one
    if optimization_calculation_id is not None:
        r = client.get('calculations/%s/cjson' % optimization_calculation_id,
                    jsonResp=False)
        cjson = r.json()
    # If we have not calculations then just use the geometry stored in molecules
    elif best_calc is None:
        r = client.get('molecules/%s/cjson' % molecule_id, jsonResp=False)
        cjson = r.json()
        # As we might be using an unoptimized structure add the optimize step
        if 'optimization' not in calculation['properties']['calculationTypes']:
            calculation['properties']['calculationTypes'].append('optimization')
    # Fetch xyz for best geometry
    else:
        optimization_calculation_id = best_calc['_id']
        r = client.get('calculations/%s/cjson' % optimization_calculation_id,
                    jsonResp=False)
        cjson = r.json()

    # If we are using an existing calculation as the input geometry record it
    if optimization_calculation_id is not None:
        props = calculation['properties']
        props['input'] = {
            'calculationId': optimization_calculation_id
        }
        calculation = client.put('calculations/%s/properties' % calculation['_id'],
                                json=props)

    oc_folder = _get_oc_folder(client)
    run_folder = client.createFolder(oc_folder['_id'],
                                    datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%f"))
    input_folder = client.createFolder(run_folder['_id'],
                                    'input')

    # Generate input file
    params = {}
    calculation_types = parse('properties.calculationTypes').find(calculation)
    if calculation_types:
        calculation_types = calculation_types[0].value

    for calculation_type in calculation_types:
        params[calculation_type] = True

    # If we have been asked to use a optimized structure make sure we
    # run the optimization if we couldn't find calculation.
    if optimize and optimization_calculation_id is None:
        params['optimization'] = True

    basis = parse('properties.basisSet.name').find(calculation)
    if basis:
        params['basis'] = basis[0].value

    functional = parse('properties.functional').find(calculation)
    if functional:
        params['functional'] = functional[0].value.lower()

    theory = parse('properties.theory').find(calculation)
    if theory:
        params['theory'] = theory[0].value.lower()

    with tempfile.TemporaryFile() as fp:
        task.taskflow.input_generator(params, cjson, fp)
        
        # Get the size of the file
        size = fp.seek(0, 2)
        fp.seek(0)
        name = 'oc.%s.in' % task.taskflow.code_label
        input_file = client.uploadFile(input_folder['_id'],  fp, name, size,
                                    parentType='folder')

    submit_template.delay(input_, cluster, run_folder, input_file, input_folder)

def _create_job_ec2(task, cluster, input_file, input_folder):
    task.taskflow.logger.info('Create %s job' % task.taskflow.code_label)
    input_name = input_file['name']

    body = {
        'name': '%s_run' % task.taskflow.code_label,
        'commands': task.taskflow.ec2_job_commands(input_name),
        'input': [
            {
              'folderId': input_folder['_id'],
              'path': '.'
            }
        ],
        'output': [],
        'params': {
            'taskFlowId': task.taskflow.id
        }
    }

    client = create_girder_client(
                task.taskflow.girder_api_url, task.taskflow.girder_token)

    job = client.post('jobs', data=json.dumps(body))
    task.taskflow.set_metadata('jobs', [job])

    return job

def _create_job_demo(task, cluster, input_file, input_folder):
    task.taskflow.logger.info('Create %s job' % task.taskflow.code_label)
    input_name = input_file['name']

    body = {
        'name': '%s_run' % task.taskflow.code_label,
        'commands': task.taskflow.demo_job_commands(input_name),
        'input': [
            {
              'folderId': input_folder['_id'],
              'path': '.'
            }
        ],
        'output': [],
        'params': {
            'taskFlowId': task.taskflow.id
        }
    }

    client = create_girder_client(
                task.taskflow.girder_api_url, task.taskflow.girder_token)

    job = client.post('jobs', data=json.dumps(body))
    task.taskflow.set_metadata('jobs', [job])

    return job



def _create_job_nersc(task, cluster, input_file, input_folder):
    task.taskflow.logger.info('Create %s job' % task.taskflow.code_label)
    input_name = input_file['name']

    body = {
        'name': '%s_run' % task.taskflow.code_label,
        'commands': task.taskflow.nersc_job_commands(input_name),
        'input': [
            {
              'folderId': input_folder['_id'],
              'path': '.'
            }
        ],
        'output': [],
        'params': {
            'taskFlowId': task.taskflow.id,
            'numberOfNodes': 1,
            'queue': 'debug',
            'constraint': 'haswell',
            'account': os.environ.get('OC_ACCOUNT')
        }
    }

    client = create_girder_client(
                task.taskflow.girder_api_url, task.taskflow.girder_token)

    job = client.post('jobs', data=json.dumps(body))
    task.taskflow.set_metadata('jobs', [job])

    return job

def _nersc(cluster):
    return cluster.get('name') in ['cori']

def _demo(cluster):
    return cluster.get('name') == 'demo_cluster'

def _create_job(task, cluster, input_file, input_folder):
    if _nersc(cluster):
        return _create_job_nersc(task, cluster, input_file, input_folder)
    elif _demo(cluster):
        return _create_job_demo(task, cluster, input_file, input_folder)
    else:
        return _create_job_ec2(task, cluster, input_file, input_folder)


@cumulus.taskflow.task
def submit_template(task, input_, cluster, run_folder, input_file, input_folder):
    job = _create_job(task, cluster, input_file, input_folder)

    girder_token = task.taskflow.girder_token
    task.taskflow.set_metadata('cluster', cluster)

    # Now download and submit job to the cluster
    task.taskflow.logger.info('Downloading input files to cluster.')
    download_job_input_folders(cluster, job,
                               girder_token=girder_token, submit=False)
    task.taskflow.logger.info('Downloading complete.')

    task.taskflow.logger.info('Submitting job %s to cluster.' % job['_id'])
    girder_token = task.taskflow.girder_token

    try:
        submit_job(cluster, job, girder_token=girder_token, monitor=False)
    except:
        import traceback
        traceback.print_exc()

    monitor_job.apply_async((cluster, job), {'girder_token': girder_token,
                                             'monitor_interval': 10},
                            link=postprocess_template.s(run_folder, input_, cluster, job))


@cumulus.taskflow.task
def postprocess_template(task, _, run_folder, input_, cluster, job):
    task.taskflow.logger.info('Uploading results from cluster')

    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    output_folder = client.createFolder(run_folder['_id'],
                                       'output')
    # Refresh state of job
    job = client.get('jobs/%s' % job['_id'])
    job['output'] = [{
        'folderId': output_folder['_id'],
        'path': '.'
    }]

    upload_job_output_to_folder(cluster, job, girder_token=task.taskflow.girder_token)

    task.taskflow.logger.info('Upload job output complete.')

    input_file_name = task.taskflow.get_metadata('inputFileName')
    input_file_name

    output_items = list(client.listItem(output_folder['_id']))
    output_filenames = [item['name'] for item in output_items]
    do_copy = task.taskflow.select_output_files(output_filenames)
    # Call to ingest the files
    for item, copy in zip(output_items, do_copy):
        if copy:
            files = list(client.listFile(item['_id']))
            if len(files) != 1:
                raise Exception('Expecting a single file under item, found: %s' + len(files))

            json_output_file_id = files[0]['_id']
            # Now call endpoint to ingest result
            body = {
                'fileId': json_output_file_id,
                'public': True
            }
            client.put('calculations/%s' % input_['calculation']['_id'], json=body)



import json
from . import avogadro

def cjson_to_xyz(cjson):
    xyz = avogadro.convert_str(json.dumps(cjson), 'cjson', 'xyz')
    # remove the first two lines in the xyz file
    # (i.e. number of atom and optional comment)
    xyz = xyz.split('\n')[2:]
    xyz = '\n  '.join(xyz)
    return xyz



import os
import jinja2
from openchemistry import OpenChemistryTaskFlow
from openchemistry.utils import cjson_to_xyz

class Psi4TaskFlow(OpenChemistryTaskFlow):

    @property
    def code_label(self):
        return 'psi4'

    @property
    def docker_image(self):
        return 'openchemistry/psi4:latest'

    def input_generator(self, params, cjson, tmp_file):
        xyz_structure = cjson_to_xyz(cjson)

        optimization = params.get('optimization', None)
        vibrational = params.get('vibrational', None)
        charge = params.get('charge', 0)
        multiplicity = params.get('multiplicity', 1)
        theory = params.get('theory', 'scf')
        functional = params.get('functional', 'b3lyp')
        basis = params.get('basis', 'cc-pvdz')

        if optimization is not None:
            task = 'optimize'
        elif vibrational is not None:
            task = 'frequency'
        else:
            task = 'energy'

        if theory.lower() == 'dft':
            _theory = functional
            reference = 'ks'
        else:
            _theory = theory
            reference = 'hf'

        if multiplicity == 1:
            reference = 'r' + reference
        else:
            reference = 'u' + reference

        context = {
            'task': task,
            'theory': _theory,
            'reference': reference,
            'charge': charge,
            'multiplicity': multiplicity,
            'basis': basis
        }

        template_path = os.path.dirname(__file__)
        jinja2_env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_path),
                                        trim_blocks=True)
        jinja2_env.get_template('oc.psi4.j2').stream(**context, xyz_structure=xyz_structure).dump(tmp_file, encoding='utf8')

    def select_output_files(self, filenames):
        do_copy = [False] * len(filenames)
        for i, file in enumerate(filenames):
            if file.endswith('.out'):
                do_copy[i] = True
        return do_copy



import os
import jinja2
from openchemistry import OpenChemistryTaskFlow
from openchemistry.utils import cjson_to_xyz

class NWChemTaskFlow(OpenChemistryTaskFlow):

    @property
    def code_label(self):
        return 'nwchem'

    @property
    def docker_image(self):
        return 'openchemistry/nwchem-json:latest'

    def input_generator(self, params, cjson, tmp_file):
        xyz_structure = cjson_to_xyz(cjson)
        template_path = os.path.dirname(__file__)
        jinja2_env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_path),
                                        trim_blocks=True)
        jinja2_env.get_template('oc.nw.j2').stream(**params, xyz_structure=xyz_structure).dump(tmp_file, encoding='utf8')

    def select_output_files(self, filenames):
        do_copy = [False] * len(filenames)
        for i, file in enumerate(filenames):
            if file.endswith('.json'):
                do_copy[i] = True
        return do_copy

    def demo_job_commands(self, input_name):
        mount_dir = '/data'
        return [
            'docker pull %s' % self.docker_image,
            'docker run -w %s -v dev_job_data:%s %s %s' % (
                os.path.join(mount_dir, '{{job._id}}'),
                mount_dir,
                self.docker_image,
                input_name
            )
        ]

    def nersc_job_commands(self, input_name):
        return [
            '/usr/bin/srun -N 1  -n 32 %s %s' % (os.environ.get('OC_NWCHEM_PATH', 'nwchem'), input_name)
        ]



import os
import requests
import re

from IPython.lib import kernel


def lookup_file(girder_client, jupyterhub_url):
    """
    Utility function to lookup the Girder file that the current notebook is running
    from.
    """
    connection_file_path = kernel.get_connection_file()
    me = girder_client.get('user/me')
    login = me['login']
    kernel_id = re.search('kernel-(.*).json', connection_file_path).group(1)

    # Authenticate with jupyterhub so we can get the cookie
    r = requests.post('%s/hub/login' % jupyterhub_url, data={'Girder-Token': girder_client.token}, allow_redirects=False)
    r.raise_for_status()
    cookies = r.cookies

    url = "%s/user/%s/api/sessions" % (jupyterhub_url, login)
    r = requests.get(url, cookies=cookies)
    r.raise_for_status()


    sessions = r.json()
    matching = [s for s in sessions if s['kernel']['id'] == kernel_id]
    path = matching[0]['path']
    name = os.path.basename(path)

    # Logout
    url = "%s/hub/logout" % jupyterhub_url
    r = requests.get(url, cookies=cookies)


    return girder_client.resourceLookup('user/%s/Private/oc/notebooks/%s/%s' % (login, path, name))




from jinja2 import Environment, BaseLoader
from girder_client import GirderClient, HttpError
import re
import os
import urllib.parse
import json
import inspect
from jsonpath_rw import parse

from .utils import lookup_file
from avogadro import io as avo_io, core

from .io.psi4 import Psi4Reader
from .io.nwchemJson import NWChemJsonReader

girder_host = os.environ.get('GIRDER_HOST')
girder_port = os.environ.get('GIRDER_PORT')
girder_scheme = os.environ.get('GIRDER_SCHEME', 'http')
girder_api_root = os.environ.get('GIRDER_API_ROOT', '/api/v1')
girder_api_key = os.environ.get('GIRDER_API_KEY')
girder_token = os.environ.get('GIRDER_TOKEN')
app_base_url = os.environ.get('APP_BASE_URL')
cluster_id = os.environ.get('CLUSTER_ID')
jupyterhub_url = os.environ.get('OC_JUPYTERHUB_URL')

if girder_host:
    girder_client = GirderClient(host=girder_host, port=girder_port,
                                 scheme=girder_scheme, apiRoot=girder_api_root)

    if girder_api_key is not None:
        girder_client.authenticate(apiKey=girder_api_key)
    elif girder_token is not None:
        girder_client.token = girder_token

    girder_file = lookup_file(girder_client, jupyterhub_url)

# TODO Need to use basis and theory
def _fetch_calculation(molecule_id, type_=None, basis=None, theory=None, functional=None, code='nwchem'):
    parameters = {
        'moleculeId': molecule_id,
        'sortByTheory': True
    }

    if type_ is not None:
        parameters['calculationType'] = type_

    if functional is not None:
        parameters['functional'] = functional

    if theory is not None:
        parameters['theory'] = theory

    if basis is not None:
        parameters['basis'] = basis

    if code is not None:
        parameters['code'] = code

    calculations = girder_client.get('calculations', parameters)

    if len(calculations) < 1:
        return None

    # Pick the "best"
    return calculations[0]

def _nersc():
    return os.environ.get('OC_SITE') == 'NERSC'

def _submit_calculation(cluster_id, pending_calculation_id, optimize, calculation_types=None, code='nwchem'):
    if cluster_id is None and not _nersc():
        # Try to get demo cluster
        params = {
            'type': 'trad'
        }
        clusters = girder_client.get('clusters', params)

        if len(clusters) > 0:
            cluster_id = clusters[0]['_id']
        else:
            raise Exception('Unable to submit calculation, no cluster configured.')

    code_params = {
        'nwchem': {
            'label': 'NWChem (version 27327)',
            'class': 'openchemistry.nwchem.NWChemTaskFlow'
        },
        'psi4': {
            'label': 'PSI4 (version 1.2.1)',
            'class': 'openchemistry.psi4.Psi4TaskFlow'
        },
        'chemml': {
            'label': 'ChemML',
            'class': 'openchemistry.chemml.ChemmlTaskFlow'
        }
    }

    if code not in code_params:
        raise Exception('Unable to submit calculation with the %s code' % code)

    code_label = code_params[code]['label']
    taskflow_class = code_params[code]['class']

    # Create the taskflow
    body = {
        'taskFlowClass': taskflow_class,
        'meta': {
            'code': code_label
        }
    }
    if calculation_types is not None:
        body['meta']['type'] = calculation_types

    taskflow = girder_client.post('taskflows', json=body)
    # Start the taskflow
    body = {
        'input': {
            'calculation': {
                '_id': pending_calculation_id
            },
            'optimize': optimize
        }
    }

    if cluster_id is not None:
        body['cluster'] = {
            '_id': cluster_id
        }
    elif _nersc():
        body['cluster'] = {
            'name': 'cori'
        }

    girder_client.put('taskflows/%s/start' % taskflow['_id'], json=body)

    # Set the pending calculation id in the meta data
    body = {
        'meta.calculationId': pending_calculation_id
    }
    girder_client.patch('taskflows/%s' % taskflow['_id'], json=body)

    return taskflow['_id']

def _fetch_taskflow_status(taskflow_id):
    r = girder_client.get('taskflows/%s/status' % taskflow_id)

    return r['status']

def _create_pending_calculation(molecule_id, type_, basis=None, theory=None, functional=None,
                                input_geometry=None, code='nwchem'):
    if not isinstance(type_, list):
        type_ = [type_]

    body = {
        'moleculeId': molecule_id,
        'cjson': None,
        'public': True,
        'properties': {
            'calculationTypes': type_,
            'pending': True,
            'code': code
        },
        'notebooks': [girder_file['_id']]
    }

    if basis is not None:
        body['properties']['basisSet'] = {
            'name': basis.lower()
        }

    if theory is not None:
        body['properties']['theory'] = theory.lower()

    if input_geometry is not None:
        body['properties']['input'] = {
            'calculationId': input_geometry
        }

    if functional is not None:
        body['properties']['functional'] = functional.lower()

    calculation = girder_client.post('calculations', json=body)

    return calculation

def _fetch_or_submit_calculation(molecule_id, type_, basis=None, theory=None, functional=None, optimize=False,
                                 input_geometry=None, code='nwchem'):
    global cluster_id
    # If a functional has been provided default theory to dft
    if theory is None and functional is not None:
        theory = 'dft'

    calculation = _fetch_calculation(molecule_id, type_, basis, theory, functional, code=code)
    taskflow_id = None

    if calculation is None:
        calculation = _create_pending_calculation(molecule_id, type_, basis,
                                                  theory, functional, input_geometry=input_geometry, code=code)
        calculation_types = parse('properties.calculationTypes').find(calculation)[0].value
        taskflow_id = _submit_calculation(cluster_id, calculation['_id'], optimize, calculation_types, code=code)
        # Patch calculation to include taskflow id
        props = calculation['properties']
        props['taskFlowId'] = taskflow_id
        calculation = girder_client.put('calculations/%s/properties' % calculation['_id'], json=props)
    else:
        # If we all ready have a calculation tag it with this notebooks id
        body = {
            'notebooks': [girder_file['_id']]
        }
        girder_client.patch('calculations/%s/notebooks' % calculation['_id'],
                            json=body)

    return calculation

def _optimize(molecule_id, basis=None, theory=None, functional=None, input_geometry=None, code='nwchem'):
    type_ = 'optimization'
    calculation =  _fetch_or_submit_calculation(molecule_id, type_, basis, theory,
                                                functional, input_geometry=input_geometry, code=code)
    pending = parse('properties.pending').find(calculation)
    if pending:
        pending = pending[0].value

    taskflow_id = parse('properties.taskFlowId').find(calculation)
    if taskflow_id:
        taskflow_id = taskflow_id[0].value
    else:
        taskflow_id = None
    calculation = CalculationResult(calculation['_id'], calculation['properties'], molecule_id)

    if pending:
        calculation = PendingCalculationResultWrapper(calculation, taskflow_id)

    return calculation

def _frequencies(molecule_id,  optimize=False, basis=None, theory=None,
                 functional=None, input_geometry=None, code='nwchem'):
    type_ = 'vibrational'
    calculation = _fetch_or_submit_calculation(molecule_id, type_, basis, theory,
                                               functional, optimize, input_geometry, code=code)
    pending = parse('properties.pending').find(calculation)
    if pending:
        pending = pending[0].value

    taskflow_id = parse('properties.taskFlowId').find(calculation)
    if taskflow_id:
        taskflow_id = taskflow_id[0].value
    else:
        taskflow_id = None
    calculation = FrequenciesCalculationResult(calculation['_id'], calculation['properties'], molecule_id)

    if pending:
        calculation = PendingCalculationResultWrapper(calculation, taskflow_id)

    return calculation

def _energy(molecule_id, optimize=False, basis=None, theory=None, functional=None, input_geometry=None, code='nwchem'):
    type_ = 'energy'
    calculation = _fetch_or_submit_calculation(molecule_id, type_, basis, theory,
                                               functional, optimize, input_geometry, code=code)
    pending = parse('properties.pending').find(calculation)
    if pending:
        pending = pending[0].value

    taskflow_id = parse('properties.taskFlowId').find(calculation)
    if taskflow_id:
        taskflow_id = taskflow_id[0].value
    else:
        taskflow_id = None
    calculation = CalculationResult(calculation['_id'], calculation['properties'], molecule_id)

    if pending:
        calculation = PendingCalculationResultWrapper(calculation, taskflow_id)

    return calculation

def _predict(molecule_id, code='chemml'):
    type_ = 'machine_learning'
    calculation = _fetch_or_submit_calculation(molecule_id, type_, code=code)
    pending = parse('properties.pending').find(calculation)
    if pending:
        pending = pending[0].value

    taskflow_id = parse('properties.taskFlowId').find(calculation)
    if taskflow_id:
        taskflow_id = taskflow_id[0].value
    else:
        taskflow_id = None
    calculation = CalculationResult(calculation['_id'], calculation['properties'], molecule_id)

    if pending:
        calculation = PendingCalculationResultWrapper(calculation, taskflow_id)

    return calculation

class Molecule(object):
    def __init__(self, cjson):
        self._cjson = cjson

    @property
    def structure(self):
        return Structure(cjson=self._cjson)

    @property
    def orbitals(self):
        return Orbitals(self._cjson)

class GirderMolecule(Molecule):
    '''
    Derived version that allows calculations to be initiated on using Girder
    '''
    def __init__(self, _id, cjson=None):
        self._id = _id
        self._cjson = cjson

    def optimize(self, basis=None, theory=None, functional=None, code='nwchem'):
        return _optimize(self._id, basis, theory, functional, code=code)

    def frequencies(self, optimize=False, basis=None, theory=None, functional=None, code='nwchem'):
        return _frequencies(self._id, optimize, basis, theory, functional, code=code)

    def energy(self, optimize=False, basis=None, theory=None, functional=None, code='nwchem'):
        return _energy(self._id, optimize, basis, theory, functional, code=code)

    def predict(self, code='chemml'):
        return _predict(self._id, code=code)

class Structure(object):

    def __init__(self, calculation_result=None, cjson=None):
        self._calculation_result = calculation_result
        self._cjson = cjson

    def show(self, style='ball-stick'):

        try:
            from .notebook import CJSON
            if self._calculation_result:
                return CJSON(self._calculation_result._cjson, vibrational=False)
            else:
                return CJSON(self._cjson, vibrational=False)
        except ImportError:
            # Outside notebook print CJSON
            print(self._calculation_result._cjson)

    def url(self, style='ball-stick'):
        url = '%s/calculations/%s' % (app_base_url.rstrip('/'), self._calculation_result._id)
        try:
            from IPython.display import Markdown
            return Markdown('[%s](%s)' % (url, url))
        except ImportError:
            # Outside notebook just print the url
            print(url)


class Frequencies(object):

    def __init__(self, calculation_result):
        self._calculation_result = calculation_result

    def show(self, mode=None, animate_modes=False, spectrum=True):
        try:
            from .notebook import CJSON
            return CJSON(self._calculation_result._cjson, structure=animate_modes,
                         animate_mode=mode)
        except ImportError:
            # Outside notebook print CJSON
            print(self.table)

    def table(self):
        return self._calculation_result._vibrational_modes

class Orbitals(object):

    def __init__(self, cjson):
        self._cjson = cjson

    # Default implementation calculate mo locally
    def _calculate_mo(self, mo):
        if isinstance(mo, str):
            mo = mo.lower()
            if mo.lower() in ['homo', 'lumo']:
                # Electron count might be saved in several places...
                path_expressions = [
                    'orbitals.electronCount',
                    'basisSet.electronCount',
                    'properties.electronCount'
                ]
                matches = []
                for expr in path_expressions:
                    matches.extend(parse(expr).find(self._cjson))
                if len(matches) > 0:
                    electron_count = matches[0].value
                else:
                    raise Exception('Unable to access electronCount')

                # The index of the first orbital is 0, so homo needs to be
                # electron_count // 2 - 1
                if mo.lower() == 'homo':
                    mo = int(electron_count / 2) - 1
                elif mo.lower() == 'lumo':
                    mo = int(electron_count / 2)
            else:
                raise ValueError('Unsupported mo: %s' % mo)

        mol = core.Molecule()
        conv = avo_io.FileFormatManager()
        conv.readString(mol, json.dumps(self._cjson), 'cjson')
        # Do some scaling of our spacing based on the size of the molecule.
        atomCount = mol.atomCount()
        spacing = 0.30
        if atomCount > 50:
            spacing = 0.5
        elif atomCount > 30:
            spacing = 0.4
        elif atomCount > 10:
            spacing = 0.33
        cube = mol.addCube()
        # Hard wiring spacing/padding for now, this could be exposed in future too.
        cube.setLimits(mol, spacing, 4)
        gaussian = core.GaussianSetTools(mol)
        gaussian.calculateMolecularOrbital(cube, mo)

        return json.loads(conv.writeString(mol, "cjson"))['cube']


    def show(self, mo='homo', iso=None):
        try:
            from .notebook import CJSON

            cjson_copy = self._cjson.copy()
            cjson_copy['cube'] = self._calculate_mo(mo)

            extra = {}
            if iso:
                extra['iso_value'] = iso

            return CJSON(cjson_copy, vibrational=False, mo=mo, **extra)
        except ImportError:
            # Outside notebook print CJSON
            print(self._cjson)


class CalculationResultOrbitals(Orbitals):

    def __init__(self, calculation_result):
        self._calculation_result = calculation_result
        super(CalculationResultOrbitals, self).__init__(calculation_result._cjson)

    def show(self, mo='homo', iso=None):
        # Save parameter to use in url
        self._last_mo = mo
        self._last_iso = iso

        return super(CalculationResultOrbitals, self).show(mo, iso)

    def url(self):
        url = '%s/calculations/%s' % (app_base_url.rstrip('/'), self._calculation_result._id)

        params = { }

        if self._last_mo is not None:
            params['mo'] = self._last_mo

        if self._last_iso is not None:
            params['iso'] = self._last_iso

        if params:
            url = '%s?%s' % (url, urllib.parse.urlencode(params))

        try:
            from IPython.display import Markdown
            return Markdown('[%s](%s)' % (url, url))
        except ImportError:
            # Outside notebook just print the url
            print(url)

    def _calculate_mo(self, mo):
        return self._calculation_result._cube(mo)['cube']

class Properties(object):

    def __init__(self, calculation_result=None, cjson=None):
        self._calculation_result = calculation_result
        self._cjson = cjson

    def show(self):

        try:
            from IPython.display import Markdown
            if self._calculation_result:
                table = self._properties_table(self._calculation_result._cjson)
            else:
                table = self._properties_table(self._cjson)
            return Markdown(table)
        except ImportError:
            # Outside notebook print CJSON
            print(self._calculation_result._cjson)

    def url(self, style='ball-stick'):
        url = '%s/calculations/%s' % (app_base_url.rstrip('/'), self._calculation_result._id)
        try:
            from IPython.display import Markdown
            return Markdown('[%s](%s)' % (url, url))
        except ImportError:
            # Outside notebook just print the url
            print(url)

    def _properties_table(self, cjson):
        import math
        table = '''### Calculated Properties
| Name | Value | Units |
|------|-------|-------|'''
        properties = cjson.get('calculatedProperties', {})

        for prop in properties.values():
            value = prop.get('value', math.nan)
            try:
                value = float(value)
            except ValueError:
                value = math.nan
            table += '\n| %s | %.2f | %s |' % (
                prop.get('label', ''),
                value,
                prop.get('units', '')
            )

        return table


class CalculationResult(object):

    def __init__(self, _id=None, properties=None, molecule_id=None):
        self._id = _id
        self._cjson_ = None
        self._vibrational_modes_ = None
        self._orbitals = None
        self._molecule_id = molecule_id
        self._properties = properties

    @property
    def _cjson(self):
        if self._cjson_ is None:
            self._cjson_ = girder_client.get('calculations/%s/cjson' % self._id)

        return self._cjson_

    @property
    def _vibrational_modes(self):
        if self._vibrational_modes_ is None:
            self._vibrational_modes_ = girder_client.get('calculations/%s/vibrationalmodes' % self._id)

        return self._vibrational_modes_

    def _cube(self, mo):
        return girder_client.get('calculations/%s/cube/%s' % (self._id, mo))

    @property
    def structure(self):
        return Structure(self)

    @property
    def orbitals(self):
        if self._orbitals is None:
            self._orbitals = CalculationResultOrbitals(self)

        return self._orbitals

    @property
    def properties(self):
        return Properties(self)


    def optimize(self, basis=None, theory=None, functional=None):
        return _optimize(self._molecule_id, basis, theory, functional, self._id)

    def frequencies(self, optimize=False, basis=None, theory=None, functional=None):
        return _frequencies(self._molecule_id, optimize, basis, theory,
                            functional, self._id)

    def energy(self, optimize=False, basis=None, theory=None, functional=None):
        return _energy(self._molecule_id, optimize, basis, theory, functional,
                       self._id)


class FrequenciesCalculationResult(CalculationResult):
    def __init__(self, _id=None, properties=None, molecule_id=None):
        super(FrequenciesCalculationResult, self).__init__(_id, properties,
                                                           molecule_id)

    @property
    def frequencies(self):
        return Frequencies(self)

class AttributeInterceptor(object):
    def __init__(self, wrapped, value, intercept_func=lambda : True):
        self._wrapped = wrapped
        self._value = value
        self._intercept_func = intercept_func


    def __getattribute__(self, name):
        # Use object's implementation to get attributes, otherwise
        # we will get recursion
        _wrapped = object.__getattribute__(self, '_wrapped')
        _value = object.__getattribute__(self, '_value')
        intercept_func = object.__getattribute__(self, '_intercept_func')

        if intercept_func() and hasattr(_wrapped, name):
            attr = object.__getattribute__(_wrapped, name)
            if inspect.ismethod(attr):
                def pending(*args, **kwargs):
                    return _value
                return pending
            else:
                return AttributeInterceptor(attr, _value, intercept_func)
        else:
            return object.__getattribute__(_wrapped, name)

class PendingCalculationResultWrapper(AttributeInterceptor):
    def __init__(self, calculation, taskflow_id=None):
        try:
            from .notebook import CalculationMonitor
            if taskflow_id is None:
                taskflow_id = calculation.properties['taskFlowId']

            table = CalculationMonitor({
                'taskFlowIds': [taskflow_id],
                'girderToken': girder_client.token
            })
        except ImportError:
            # Outside notebook just print message
            table = 'Pending calculations .... '

        # Only intercept when the taskflow is not complete
        def intercept():
            return _fetch_taskflow_status(taskflow_id) != 'complete'

        super(PendingCalculationResultWrapper, self).__init__(calculation,
                                                              table, intercept)

class Reaction(object):
    def __init__(self, equation):
        self._equation = equation

        (self._reactants, self._products) = equation.split('=>')
        self._reactants = [x.strip() for x in self._reactants.split('+')]
        self._products = [x.strip() for x in self._products.split('+')]

    @property
    def reactants(self):
        return self._reactants

    @property
    def products(self):
        return self._products

    @property
    def equation(self):
        return '%s => %s' % (' + '.join(self.reactants), ' + '.join(self.products))

    def _fetch_free_energy(self, formula, basis=None, theory=None, functional=None):
        """
        :return A tuple containing the total energy and zero point energy.
        """

        # First fetch the molecule using the formula
        params = {
            'formula': formula
        }
        mol = girder_client.get('molecules/search', parameters=params)

        if len(mol) < 1:
            raise Exception('No molecules found for formula \'%s\'' % formula)

        # TODO Might we get more than one molecule with the same formula?

        # Now fetch the calculations, TODO what types should we select
        calculation = _fetch_or_submit_calculation(mol[0]['_id'], ['vibrational',
                                                                   'energy'],
                                                   basis, theory, functional)

        pending = parse('properties.pending').find(calculation)
        if pending:
            pending = pending[0].value

        if pending:
            taskflow_id = parse('properties.taskFlowId').find(calculation)
            taskflow_id = taskflow_id[0].value
            return CalculationResult(calculation['_id'], calculation['properties'])

        calcs = parse('properties.calculations').find(calculation)
        if not calcs:
            raise Exception('No calculations found for \'%s\'' % formula)

        calcs = calcs[0].value

        # TODO for now just select the first, which calculations should we
        # favor? For now just search for the first that has both energies
        selected_calc = None
        for calc in calcs:
            if 'totalEnergy' in calc and 'zeroPointEnergyCorrection' in calc:
                selected_calc = calc
                break

        return (selected_calc['totalEnergy'], selected_calc['zeroPointEnergyCorrection'])

    def free_energy(self, basis=None, theory=None, functional=None):

        def _sum(formulas):
            pending_calculations = []
            energy = 0
            for formula in formulas:
                free_energy = self._fetch_free_energy(formula, basis, theory, functional)

                if isinstance(free_energy, CalculationResult):
                    pending_calculations.append(free_energy)
                else:
                    (total_energy, zero_point_energy) = free_energy
                    energy += total_energy['value'] + zero_point_energy['value']

            if len(pending_calculations) == 0:
                return energy
            else:
                return pending_calculations

        reactants_energy_total = _sum(self.reactants)
        products_energy_total = _sum(self.products)

        if isinstance(reactants_energy_total, list) or isinstance(products_energy_total, list):
            pending_calculations = []
            if isinstance(reactants_energy_total, list):
                pending_calculations += reactants_energy_total

            if isinstance(products_energy_total, list):
                pending_calculations += products_energy_total

            return pending_calculations

        free_energy = products_energy_total - reactants_energy_total
        # Convert to kJ/mol
        free_energy = free_energy * 2625.5

        return free_energy

_inchi_key_regex = re.compile("^([0-9A-Z\-]+)$")

def _is_inchi_key(identifier):
    return len(identifier) == 27 and identifier[25] == '-' and \
        _inchi_key_regex.match(identifier)

def _find_using_cactus(identifier):
    params = {
        'cactus': identifier

    }
    molecule = girder_client.get('molecules/search', parameters=params)

    # Just pick the first
    if len(molecule) > 0:
        molecule = molecule[0]
        return GirderMolecule(molecule['_id'], molecule['cjson'])
    else:
        return None

def find_structure(identifier, basis=None, theory=None, functional=None, code='nwchem'):
    is_calc_query = (basis is not None or theory is not None
                     or functional is not None)

    # InChiKey?
    if _is_inchi_key(identifier):
        try:
            molecule = girder_client.get('molecules/inchikey/%s' % identifier)

            # Are we search for a specific calculation?
            if is_calc_query:
                # Look for optimization calculation
                cal = _fetch_calculation(molecule['_id'], 'optimization',
                                         basis, theory, functional, code=code)

                if cal is not None:
                    # TODO We should probably pass in the full calculation
                    # so we don't have to fetch it again.
                    return CalculationResult(cal['_id'])
                else:
                    return None
            else:
                return GirderMolecule(molecule['_id'], molecule['cjson'])
        except HttpError as ex:
            if ex.status == 404:
                # Use cactus to try a lookup the structure
                molecule = _find_using_cactus(identifier)
            else:
                raise

    # If we have been provided basis, theory or functional and we haven't found
    # a calculation, then we are done.
    if is_calc_query:
        return None

    # Try cactus
    molecule = _find_using_cactus(identifier)


    if not molecule:
        raise Exception('No molecules found matching identifier: \'%s\'' % identifier)

    return molecule


def setup_reaction(equation):
    return Reaction(equation)

def compose_equation(equation, **vars):
    equation = Environment(loader=BaseLoader()).from_string(equation)

    return equation.render(**vars)

def show_free_energies(reactions, basis=None, theory=None, functional=None):
    free_energy_chart_data = {
        'freeEnergy': [],
        'reaction': []
    }

    pending_calculations = []
    for reaction in reactions:
        equation = reaction.equation
        free_energy = reaction.free_energy(basis, theory, functional)

        if isinstance(free_energy, list):
            pending_calculations += free_energy

        free_energy_chart_data['reaction'].append(equation)
        free_energy_chart_data['freeEnergy'].append(free_energy)

    if pending_calculations:
        taskflow_ids = [ cal.properties['taskFlowId'] for cal in pending_calculations]
        # Remove duplicates
        taskflow_ids = list(set(taskflow_ids))

        try:
            from .notebook import CalculationMonitor
            table = CalculationMonitor({
                    'taskFlowIds': taskflow_ids,
                    'girderToken': girder_client.token
                })
        except ImportError:
            # Outside notebook just print message
            table = 'Pending calculations .... '

        return table;

    try:
        from .notebook import FreeEnergy

        return FreeEnergy(free_energy_chart_data)
    except ImportError:
        # Outside notebook print the data
        print(free_energy_chart_data)

def load(cjson):
    return Molecule(cjson)

